---
id: packages
title: Packages
sidebar_label: Packages
---

# Packages

Before contributing a new package, it will go through a proposal process to be accepted by the Steering Committee.

## Folder structure

### Tracked folders

Always include an `src` folder at the root level of the package. The `src` folder should include all of the un-minified source code and test files for the project. Place test files in the same folder as the file they are testing and name it using the `*.spec.ts` or `*.spec.tsx` pattern.

When it is necessary to create a testing folder, it should be named `__tests__`.

For TypeScript projects, export all relevant files from the `index.ts` or `index.tsx` file in the `src` folder so that, when importing, only the folder name is needed. Keep the folder structure as flat as possible.

When a testing application is needed, it should be created in an `app` directory at the root of the package.

![Example of tracked folder structure with fast-package at the top and __tests__, app and src inside.](https://res.cloudinary.com/fast-dna/image/upload/v1546297640/creating_packages_tracked_folders.webp "")
*Example structure for tracked folders.*

### Untracked folders

Untracked folders are folders included in the `.gitignore` and generated by an automated process.

These may include:

* A `node_modules` folder for npm dependencies -- generated by the `npm i` command.
* A `dist` folder for the exported code -- generated by compiling TypeScript and specified by the `tsconfig.json`.
* A `www` folder for exported compiled applications if creating an `app` folder -- generated by `webpack` and specified by the `webpack.config.js`.
* A `coverage` folder for test coverage reporting.

## package.json

All packages use npm and are published to the [npm registry](https://www.npmjs.com/).

### Side effects

All packages are expected to be side-effect free. Adding a `"sideEffects": false` flag to the package.json indicates to tools like webpack that the package's modules have no side effects (on evaluation) and only expose exports. This allows these tools to optimize re-exports through tree-shaking and dead code removal.

### Scripts

Unless specified, scripts are defined in the `package.json` scripts property so they can be run with `npm run [script-name]`.

The following scripts are required:

* `npm run build` builds the exported package.
* `npm run clean:dist` cleans the distribution folder.
* `npm run prepare` cleans the distribution folder and builds the exported package.
* `npm run prettier` ensures code is formatted correctly (runs on a global repository githook).
* `npm run prettier:diff` provides a hook to ensure code formatting matches expectations on the build server.
* `npm run test` runs all tests.
* `npm run eslint` runs eslint.

Additional scripts and common use cases include:

* `npm start` starts the webpack-dev-server.
* `npm run eslint:fix` runs eslint with the fix option.
* `npm run watch` watches the files and builds the exported package when making changes to the source files.

## .npmignore

The `.npmignore` file removes:

* All testinging files; usually by targetting the `__tests__`, `*.spec.*` and `*.test.*` patterns.
* All unminified code; usually by targetting the `src` folder.
* The `babel.config.js` if it has been added to the package.

## Webpack

A Webpack application can be created using `webpack` and `webpack-dev-server` for testing and/or documentation purposes. If it is necessary to publish a webpack application that serves as a documentation site to a live site, a production webpack build should be made available.

Applications using `webpack-dev-server` can also serve as a means for testing using E2E testing frameworks.

## Typescript

Use TypeScript for packages that include JavaScript and make sure to add `tsconfig.json` and `.eslintrc.js` files.

Published packages must be in ES6 module format to support tree-shaking.

## README format

```markdown
# [packageName] <!-- e.g. FAST Animation -->

<!-- format: A[n] [what: e.g. library,library of components] for [what does it do?]. -->
[packageDescription]

## Benefits

<!-- List the key benefits of the package -->
* [packageBenefit1]
* [packageBenefit2]

## Package status
<!-- If package is not a proposal or experimental status, delete this section -->

<!-- Delete one to show the status: -->
:soon: **Proposal: This package contains code and APIs may change and have gaps in stability, testing and documentation. Your feedback is encouraged.**

:sos: **Experimental: This package contains code and APIs are likely to change, may not be stable, and may be missing tests and documentation.**

## Usage

### Installation

[installationGuidance]

### Getting started
<!-- When possible, provide a brief overview of how to use the package. This should be the minimum work required to get a meaningful understanding of how the package works -->

[gettingStartedGuidance]

### Properties/Configuration/etc.
<!-- Describe in detail, how to use the package. This includes APIs, configurations, ... -->

[detailedGuidance]

## Examples
<!-- If possible, show one or more examples. -->

[Examples]

## Exports
<!-- If applicable, describe the package exports and if they require additional configuration, for example because they are exported as ES6 and therefore need babel or some other method of transpiling to ES2015. -->

[Exports]

## Additional resources

<!-- If there are additional resources, include them here. -->
[Resource 1](https://)
[Resource 2](https://)
```